"""Formatting SPF records helpers"""

import logging

from netaddr import AddrFormatError, IPNetwork, IPSet

from ee_spf_flattener.config import config
from ee_spf_flattener.const import SPF_ALL_POLICIES

MAX_TXT_RECORD_VALUE_LENGTH = 255
MAX_TXT_RECORD_VALUES_LENGTH = 4000

LOG = logging.getLogger(__name__)


def compute_spf_records(
    ips, all_policy=None, record_name_format=None, max_value_length=None, max_values_length=None
):
    """
    Compute SPF records form address IP list to included as a dict

    Note: Length of TXT record value is limited to 255 characters (double-quoted, quotes not
    included in this limit), but TXT could contain multiple string-separated double-quoted values
    with a total length of 4000 characters. So, we have to split result in multiple SPF records,
    each record if a concatenation of string-separated double-quoted values with a limit of 255
    characters and the record total length could not exceeded 4000 characters and each SPF record
    include the next one (include:[next record name]).
    """
    record_name_format = record_name_format or config.get("flattening.spf_record_name_format")
    base_record = "v=spf1"
    ctx = {
        "records": {},
        "base_record": base_record,
        "record_name_format": record_name_format,
        "max_value_length": (
            max_value_length or config.get("flattening.spf_record_max_value_length")
        ),
        "max_values_length": (
            max_values_length or config.get("flattening.spf_record_max_values_length")
        ),
        "current_record_values": [],
        "current_record_value": base_record,
        "current_record_name": record_name_format.format(""),
        "next_record_name": record_name_format.format(1),
        "include_next_record_format": "include:{}",
    }

    def add(ctx, ip):
        """Add an IP to the result managing the record & record's values limits"""
        # Max values length will be exceeded with this IP?
        if (
            len(
                " ".join(
                    ctx["current_record_values"]
                    + [
                        f'"{ctx["current_record_value"]} {ip} '
                        f'{ctx["include_next_record_format"].format(ctx["next_record_name"])}"'
                    ]
                )
            )
            > ctx["max_values_length"]
        ):
            include_next_record = ctx["include_next_record_format"].format(ctx["next_record_name"])
            record = f'"{ctx["current_record_value"]} {include_next_record}"'
            next_record = ctx["base_record"]

            # Check if the last record value exceeded the value length limit:
            if len(record) > ctx["max_value_length"]:
                # Split record's tokens and put exceeded tokens in the next record
                keep_tokens = []
                next_record_tokens = []
                tokens = record[1:-1].split()
                prefix = "" if tokens[0] == base_record else " "
                for token in tokens[0:-1]:
                    if (
                        next_record_tokens
                        or len(f'"{prefix}{" ".join(keep_tokens + [token, include_next_record])}"')
                        > ctx["max_value_length"]
                    ):
                        next_record_tokens.append(token)
                    else:
                        keep_tokens.append(token)
                record = f'"{prefix}{" ".join(keep_tokens + [include_next_record])}"'
                next_record = f"{next_record} {' '.join(next_record_tokens)}"
                LOG.debug(
                    "Add record %s (%s) with a total length of %d "
                    "and put exceeded tokens in the next record ('%s')",
                    ctx["current_record_name"],
                    ", ".join(
                        [
                            f"'{value}' (length: {len(value)})"
                            for value in (ctx["current_record_values"] + [record])
                        ]
                    ),
                    len(" ".join(ctx["current_record_values"] + [record])),
                    "', '".join(next_record_tokens),
                )
            else:
                LOG.debug(
                    "Add record %s (%s) with a total length of %d",
                    ctx["current_record_name"],
                    ", ".join(
                        [
                            f"'{value}' (length: {len(value)})"
                            for value in (ctx["current_record_values"] + [record])
                        ]
                    ),
                    len(" ".join(ctx["current_record_values"] + [record])),
                )

            ctx["current_record_values"].append(record)
            ctx["records"][ctx["current_record_name"]] = " ".join(ctx["current_record_values"])
            ctx["current_record_value"] = next_record
            ctx["current_record_values"] = []
            ctx["current_record_name"] = ctx["record_name_format"].format(len(ctx["records"]))
            ctx["next_record_name"] = ctx["record_name_format"].format(len(ctx["records"]) + 1)

        # Max value length will be exceeded with this IP?
        if len(f"{ctx['current_record_value']} {ip}") > ctx["max_value_length"]:
            ctx["current_record_values"].append(f'"{ctx["current_record_value"]}"')
            ctx["current_record_value"] = ""

        ctx["current_record_value"] += f" {ip}"

    # Add all IPs and the domain ALL policy
    for ip in ips:
        add(ctx, ip)
    add(ctx, SPF_ALL_POLICIES[all_policy or config.get("flattening.spf_all_policy")])

    # Add the current record value to the record (if not empty)
    if ctx["current_record_value"]:
        ctx["current_record_values"].append(f'"{ctx["current_record_value"]}"')

    # Add the current record to the result (if not empty)
    if ctx["current_record_values"]:
        ctx["records"][ctx["current_record_name"]] = " ".join(ctx["current_record_values"])

    return ctx["records"]


def format_spf_record(name, value, ttl=None):
    """Format an SPF record"""
    # pylint: disable=consider-using-f-string
    return f'{name}{" {}".format(ttl) if ttl else ""} IN TXT {value}'


def ips_to_spf_strings(ips):
    """Convert IP list to SPF entry as string"""
    other_tokens = []
    for ip in ips:
        try:
            IPNetwork(ip)
        except AddrFormatError:
            other_tokens.append(ip)
    for token in other_tokens:
        ips.remove(token)
    ips = [str(i) for i in IPSet(ips).iter_cidrs()]
    ips = ["ip6:" + ip if ":" in ip else "ip4:" + ip.replace("/32", "") for ip in ips]
    return ips + other_tokens
