"""DNS name resolution"""
import logging

import dns
import dns.resolver

LOG = logging.getLogger(__name__)


class Resolver(dns.resolver.Resolver):
    """Custom DNS name resolver that allow to route query on based on query domain"""

    custom_domain = None
    custom_domain_nameservers = None
    custom_domain_resolver = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self, config):
        """Initialize from configuration"""
        custom_nameservers = config.get("flattening.name_servers")
        if custom_nameservers:
            LOG.debug("Use specified DNS resolvers (%s)", ", ".join(custom_nameservers))
            self.nameservers = custom_nameservers
        else:
            LOG.debug("Use system default DNS resolvers")

        self.custom_domain = dns.name.from_text(config.get("flattening.domain"))
        self.custom_domain_nameservers = config.get("flattening.domain_name_servers")
        if not self.custom_domain_nameservers:
            LOG.debug(
                "No name servers provided for domain %s, try to autodetect",
                self.custom_domain.to_text(),
            )
            self.custom_domain_nameservers = list(self.name2ipaddress(self.custom_domain, "NS"))
            assert (
                self.custom_domain_nameservers
            ), f"Fail to autodetect name servers for domain {self.custom_domain}"
            LOG.debug(
                "Detected name servers for domain %s: %s",
                self.custom_domain.to_text(),
                ", ".join(self.custom_domain_nameservers),
            )
        self.custom_domain_resolver = dns.resolver.Resolver()
        self.custom_domain_resolver.nameservers = self.custom_domain_nameservers

    def name2ipaddress(self, rrname, rrtype):
        """Get IP addresses for specified record name"""
        rrname = rrname if isinstance(rrname, dns.name.Name) else dns.name.from_text(rrname)
        try:
            LOG.debug(
                "Looking for IP addresses for %s records of %s",
                rrtype.upper(),
                rrname.to_text(),
            )
            answers = self.query(rrname, rrtype)
        except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
            LOG.error(
                "Fail to resolve %s records for %s: does not exists (resolvers: %s)",
                rrtype.upper(),
                rrname.to_text(),
                ", ".join(self.nameservers) if self.nameservers else "system default",
            )
        except Exception:  # pylint: disable=broad-exception-caught
            LOG.exception("Fail to resolve %s records for %s", rrtype.upper(), rrname.to_text())
        else:
            for answer in answers:
                if rrtype.lower() in ["cname", "ns"]:
                    try:
                        LOG.debug(
                            "Looking for IP addresses for %s %s record target for %s",
                            rrname.to_text(),
                            rrtype.upper(),
                            answer.target,
                        )
                        for ip in self.query(answer.target):
                            yield ip.address
                    except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
                        LOG.warning(e)
                elif rrtype.lower() == "mx":
                    LOG.debug(
                        "Looking for IP addresses for %s %s record target for %s",
                        rrname.to_text(),
                        rrtype.upper(),
                        answer.exchange,
                    )
                    try:
                        for ip in self.query(answer.exchange, "a"):
                            yield ip.address
                    except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
                        LOG.warning(e)
                else:
                    yield answer.address

    def query(self, qname, *args, rdtype=None, **kwargs):
        """Query nameservers to find the answer to the question."""
        rdtype = rdtype if rdtype else "A"
        qname = qname if isinstance(qname, dns.name.Name) else dns.name.from_text(qname)
        if self.custom_domain_resolver and qname.is_subdomain(self.custom_domain):
            LOG.debug(
                "Query on %s record for %s: use %s domain name servers",
                rdtype.upper(),
                qname.to_text(),
                self.custom_domain.to_text(),
            )
            return self.custom_domain_resolver.query(qname, *args, **kwargs)
        LOG.debug(
            "Query on %s record for %s: use default name servers",
            rdtype.upper(),
            qname.to_text(),
        )
        return super().query(qname, *args, **kwargs)
