"""Test subcommand"""

import logging

from dns.resolver import NXDOMAIN
from netaddr import IPAddress, IPNetwork
from netaddr.core import AddrFormatError

from ee_spf_flattener.config import config

LOG = logging.getLogger(__name__)


def test_ip(ip_to_test, domain_ips_or_networks):
    """Test if a IP address is listed in the domain SPF hosts"""
    for ip_or_network in domain_ips_or_networks:
        if ip_or_network.version != ip_to_test.version:
            continue
        if ip_to_test == ip_or_network:
            LOG.info("The IP address %s is explicitly listed in the domain's SPF hosts", ip_to_test)
            return True
        if isinstance(ip_or_network, IPNetwork) and ip_to_test in ip_or_network:
            LOG.info(
                "The IP address %s is inside the IP network %s listed in the domain's SPF networks",
                ip_to_test,
                ip_or_network,
            )
            return True
    LOG.warning("The IP address %s is not listed in the domain SPF hosts", ip_to_test)
    return False


def test_domain_name_or_ips(domain_names_or_ips, domain_spf_tokens, resolver=None):
    """Test a domain name or an IP address against SPF records"""
    domain_ips_or_networks = set()
    for token in domain_spf_tokens:
        if isinstance(token, (IPAddress, IPNetwork)):
            domain_ips_or_networks.add(token)
        else:
            LOG.warning("Unsupported %s token found in the domain's SPF record", token)

    ips_to_test = set()
    for domain_name_or_ip in domain_names_or_ips:
        try:
            ips_to_test.add(IPAddress(domain_name_or_ip))
        except AddrFormatError:
            resolver = resolver or config.resolver
            try:
                answer = resolver.resolve(domain_name_or_ip)
                LOG.info(
                    "Domain name %s resolved as %s",
                    domain_name_or_ip,
                    ", ".join([rrset.address for rrset in answer.rrset]),
                )
                for rrset in answer.rrset:
                    ips_to_test.add(IPAddress(rrset.address))
            except NXDOMAIN:
                LOG.error("Unknown domain name %s", domain_name_or_ip)
                return False
    invalid_ips = []
    for ip_to_test in ips_to_test:
        if not test_ip(ip_to_test, domain_ips_or_networks):
            invalid_ips.append(ip_to_test)
    if invalid_ips:
        LOG.warning("%d invalid IP addresses found", len(invalid_ips))
        return False
    LOG.info("All specified domain names or IP addresses are allowed by the domain's SPF record")
    return True
