"""SPF records updater"""
import json
import logging

import requests
from requests.exceptions import HTTPError

try:
    from requests.exceptions import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

from ee_spf_flattener.config import config

LOG = logging.getLogger(__name__)


class BaseUpdater:
    """Base class to implement an SPF record updater"""

    def initialize(self):
        """Initialize updater"""

    @property
    def record_name_format(self):
        """Retrieve configured SPF records name format"""
        return config.get("flattening.spf_record_name_format", required=True)

    def get_record_name(self, idx):
        """Compute and retrieve SPF record name"""
        return self.record_name_format.format(idx if idx else "")

    @staticmethod
    def get_records_domain(records):
        """Compute the domain name from the first SPF record name"""
        domain = ".".join(next(iter(records)).split(".")[1:])
        LOG.debug("Detected domain: %s", domain)
        return domain

    def update_spf_records(self, records):
        """Update SPF records"""
        raise NotImplementedError()


class RESTAPIUpdater(BaseUpdater):
    """SPF records updater via a REST API"""

    # pylint: disable=unused-private-member
    def _call_api(
        self,
        url,
        method=None,
        data=None,
        headers=None,
        decode_json=True,
        return_error=False,
        log_error=True,
    ):
        """Call powerdns API"""
        method = method if method else "GET"
        headers = headers if headers else {}

        response = requests.request(
            method,
            url,
            headers=headers,
            data=json.dumps(data) if data else None,
            timeout=config.get("update.timeout"),
        )
        try:
            response.raise_for_status()
            return response.json() if decode_json else response.text
        except JSONDecodeError as err:
            LOG.error(
                "Error occurred decoding API JSON result on %s request to %s: %s\n%s",
                method,
                url,
                err,
                response.text,
            )
        except HTTPError as err:
            if return_error:
                return response
            if log_error:
                LOG.error(
                    "Error occurred calling API on %s request to %s:%s\n%s",
                    method,
                    url,
                    err,
                    response.text,
                )
        return False

    def set_record(self, record, value, record_type=None, ttl=None):
        """
        Set the value of one record (replace all existing values)

        :param record: The record name to set
        :param value: The record value
        :param record_type: The record type to set (optional, default: TXT)
        :param ttl: The TTL of the record (optional)
        """
        raise NotImplementedError()

    def delete_record(self, record, record_type=None):
        """
        Delete all values of one record

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        raise NotImplementedError()

    def record_exists(self, record, record_type=None):
        """
        Check if one record exists

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        raise NotImplementedError()

    def update_spf_records(self, records):
        """Update SPF records by send an NS update query"""
        error = False
        for name, value in records.items():
            if self.set_record(name, value):
                LOG.debug("Record %s updated", name)
            else:
                LOG.error("Fail to update record %s with value '%s'", name, value)
                error = True

        LOG.info("Lookup for old SPF record need to be cleaned")
        count = 0
        while True:
            name = self.get_record_name(count)
            count += 1
            if name in records:
                continue
            if not self.record_exists(name, record_type="TXT"):
                # No more old SPF record to clean, stop
                break
            LOG.info("An old %s SPF record detected, purge it", name)
            if self.delete_record(name, record_type="TXT"):
                LOG.debug("Record %s deleted", name)
            else:
                LOG.error("Fail to delete record %s", name)
                error = True
        if error:
            LOG.info("Some error occurred updating SPF records")
            return False
        LOG.info("SPF records updated")
        return True


#
# Helpers functions
#
def get_fqdn(name):
    """Compute FQDN by adding the leading dot if not present"""
    return f"{name}." if not name.endswith(".") else name


def record2domain(record):
    """Retrieve domain name from a record name"""
    return ".".join(record.split(".")[1:])
