"""
Gandi SPF records updater
Note: using Gandi LiveDNS API, see https://api.gandi.net/docs/livedns/
"""
import datetime
import logging

from requests.exceptions import HTTPError

try:
    from requests.exceptions import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

from ee_spf_flattener.config import config
from ee_spf_flattener.updater import RESTAPIUpdater, record2domain

LOG = logging.getLogger(__name__)


class Updater(RESTAPIUpdater):
    """SPF records updater"""

    api_url = None
    api_key = None
    auth_api_url = None
    warning_api_key_expiration_delay = None

    def initialize(self):
        """Initialize updater"""
        if not self.api_url:
            self.api_url = config.get("gandi.api_url", required=True)
            self.api_key = config.get("gandi.api_key", required=True)
            self.auth_api_url = config.get("gandi.auth_api_url", required=True)
            self.warning_api_key_expiration_delay = config.get(
                "gandi.warning_api_key_expiration_delay"
            )
            self._check_api_key_expiration()

    def call_api(self, suburl, headers=None, auth_api=False, **kwargs):
        """Call Gandi API"""
        headers = headers if headers else {}
        headers["authorization"] = f"Bearer {self.api_key}"
        return self._call_api(
            f"{self.api_url if not auth_api else self.auth_api_url}/{suburl}",
            headers=headers,
            **kwargs,
        )

    @staticmethod
    def record_url(record, record_type=None):
        """Compute one record URL"""
        url = f"livedns/domains/{record2domain(record)}/records/{record.split('.', 1)[0]}"
        if record_type:
            url += f"/{record_type}"
        return url

    def set_record(self, record, value, record_type=None, ttl=None):
        """
        Set the value of one record (replace all existing values)

        :param record: The record name to set
        :param value: The record value
        :param record_type: The record type to set (optional, default: TXT)
        :param ttl: The TTL of the record (optional)
        """
        response = self.call_api(
            self.record_url(record),
            method="PUT",
            data={
                "items": [
                    {
                        "rrset_type": record_type.upper() if record_type else "TXT",
                        "rrset_values": [value],
                        "rrset_ttl": ttl if ttl is not None else config.get("flattening.ttl"),
                    }
                ],
            },
            headers={"content-type": "application/json"},
        )
        return bool(response)

    def delete_record(self, record, record_type=None):
        """
        Delete all values of one record

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        response = self.call_api(
            self.record_url(record, record_type=record_type),
            method="DELETE",
            decode_json=False,
            return_error=True,
        )
        if isinstance(response, HTTPError):
            # Error returned: allow only 404 error on not found DNS record
            try:
                response = response.json()
                return (
                    response.response.get("object") == "dns-record"
                    and response.get("response") == 404
                    and response.get("cause") == "Not Found"
                )
            except JSONDecodeError as err:
                LOG.error(
                    "Error occurred decoding Gandi API JSON error on deleting record %s: %s\n%s",
                    record,
                    err,
                    response.response.text,
                )
                return False
        elif isinstance(response, str):
            # Empty string returned on success
            return response == ""
        return False

    def record_exists(self, record, record_type=None):
        """
        Check if one record exists

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        response = self.call_api(self.record_url(record, record_type=record_type), log_error=False)
        return bool(response)

    def _check_api_key_expiration(self):
        """Check API key expiration"""
        if not self.warning_api_key_expiration_delay:
            return
        try:
            response = self.call_api("tokeninfo", auth_api=True)
            expiration_delay = datetime.timedelta(seconds=response["expires_in"])
            expiration_date = datetime.datetime.now() + expiration_delay
        except Exception:  # pylint: disable=broad-exception-caught
            LOG.warning("Fail to retrieve Gandi API key expiration date")
            return
        LOG.log(
            logging.WARNING
            if expiration_delay.days < self.warning_api_key_expiration_delay
            else logging.DEBUG,
            "Gandi API key will expire in %d day(s) (%s)",
            expiration_delay.days,
            expiration_date.strftime("%Y/%m/%d, %T"),
        )
