"""
PowerDNS SPF records updater
Note: using PowerDNS HTTP API, see https://doc.powerdns.com/authoritative/http-api
"""
import logging

from ee_spf_flattener.config import config
from ee_spf_flattener.updater import RESTAPIUpdater, get_fqdn, record2domain

LOG = logging.getLogger(__name__)


class Updater(RESTAPIUpdater):
    """SPF records updater"""

    api_url = None
    api_key = None
    api_server_id = None

    def initialize(self):
        """Initialize updater"""
        if not self.api_url:
            self.api_url = config.get("powerdns.api_url", required=True)
            self.api_key = config.get("powerdns.api_key", required=True)
            self.api_server_id = config.get("powerdns.api_server_id", required=True)

    def call_api(self, suburl, **kwargs):
        """Call PowerDNS API"""
        return self._call_api(
            f"{self.api_url}/{suburl}",
            headers={
                "Accept": "application/json",
                "X-API-Key": self.api_key,
            },
            **kwargs,
        )

    def zone_url(self, domain=None, record=None):
        """Compute one zone URL"""
        if domain is None:
            assert record, "You must provide domain or record name"
            domain = record2domain(record)
        return f"servers/{self.api_server_id}/zones/{get_fqdn(domain)}"

    def set_record(self, record, value, record_type=None, ttl=None):
        """
        Set the value of one record (replace all existing values)

        :param record: The record name to set
        :param value: The record value
        :param record_type: The record type to set (optional, default: TXT)
        :param ttl: The TTL of the record (optional)
        """
        return (
            self.call_api(
                self.zone_url(record=record),
                method="PATCH",
                data={
                    "rrsets": [
                        {
                            "name": get_fqdn(record),
                            "type": record_type.upper() if record_type else "TXT",
                            "ttl": ttl if ttl is not None else config.get("flattening.ttl"),
                            "changetype": "REPLACE",
                            "records": [{"content": value, "disabled": False}],
                        },
                    ],
                },
                decode_json=False,
            )
            is not False
        )

    def delete_record(self, record, record_type=None):
        """
        Delete all values of one record

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        return (
            self.call_api(
                self.zone_url(record=record),
                method="PATCH",
                data={
                    "rrsets": [
                        {
                            "name": get_fqdn(record),
                            "type": record_type.upper() if record_type else "TXT",
                            "changetype": "DELETE",
                        },
                    ],
                },
                decode_json=False,
            )
            is not False
        )

    def record_exists(self, record, record_type=None):
        """
        Check if one record exists

        :param record: The record name to set
        :param record_type: The record type to set (optional, default: TXT)
        """
        response = self.call_api(self.zone_url(record=record))
        if not isinstance(response, dict):
            return False
        fqdn = get_fqdn(record).lower()
        record_type = record_type.upper() if record_type else "TXT"
        try:
            for rrset in response["rrsets"]:
                if rrset["name"].lower() == fqdn and rrset["type"] == record_type:
                    LOG.debug("record_exists(%s, %s): record found: %s", record, record_type, rrset)
                    return True
        except KeyError:
            LOG.exception(
                "Invalid API return getting zone %s, can't determine if %s record %s exists, "
                "assume is doesn't exists",
                record2domain(record),
                record_type,
                record,
            )
        return False
